/* $Id: isdnif.h,v 1.5 1995/04/23 13:45:34 fritz Exp fritz $
 *
 * ISDN-Driver for Linux (linklevel-module)
 *
 * Definition of the interface between the linklevel- and lowlevel-routines.
 *
 * Copyright 1994,95 by Fritz Elfert (fritz@wuemaus.franken.de)
 * Copyright 1995 Thinking Objects Software GmbH Wuerzburg
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA. 
 *
 * $Log: isdnif.h,v $
 * Revision 1.5  1995/04/23  13:45:34  fritz
 * Increased size of num.
 *
 * Revision 1.4  1995/03/25  23:40:08  fritz
 * Added more constants for Protocol-selection.
 *
 * Revision 1.3  1995/02/01  11:02:16  fritz
 * Added CAUSE-Message
 *
 * Revision 1.2  1995/01/29  23:31:28  fritz
 * Added feature-flags.
 *
 * Revision 1.1  1995/01/09  07:36:55  fritz
 * Initial revision
 *
 */
#ifndef isdnif_h
#define isdnif_h

/*
 * Values for Layer-2-protocol-selection
 */
#define ISDN_PROTO_L2_X75I   0   /* X75/LAPB with I-Frames      */
#define ISDN_PROTO_L2_X75UI  1   /* X75/LAPB with UI-Frames     */
#define ISDN_PROTO_L2_X75BUI 2   /* X75/LAPB with UI-Frames     */
#define ISDN_PROTO_L2_HDLC   3   /* HDLC                        */

/*
 * Values for Layer-3-protocol-selection
 */
#define ISDN_PROTO_L3_TRANS  0   /* Transparent                 */

#ifdef __KERNEL__
/*
 * Commands from linklevel to lowlevel
 *
 */
#define ISDN_CMD_IOCTL   0       /* Perform ioctl                         */
#define ISDN_CMD_DIAL    1       /* Dial out                              */
#define ISDN_CMD_ACCEPTD 2       /* Accept an incoming call on D-Chan.    */
#define ISDN_CMD_ACCEPTB 3       /* Request B-Channel connect.            */
#define ISDN_CMD_HANGUP  4       /* Hangup                                */
#define ISDN_CMD_CLREAZ  5       /* Clear EAZ(s) of channel               */
#define ISDN_CMD_SETEAZ  6       /* Set EAZ(s) of channel                 */
#define ISDN_CMD_GETEAZ  7       /* Get EAZ(s) of channel                 */
#define ISDN_CMD_SETSIL  8       /* Set Service-Indicator-List of channel */
#define ISDN_CMD_GETSIL  9       /* Get Service-Indicator-List of channel */
#define ISDN_CMD_SETL2  10       /* Set B-Chan. Layer2-Parameter          */
#define ISDN_CMD_GETL2  11       /* Get B-Chan. Layer2-Parameter          */
#define ISDN_CMD_SETL3  12       /* Set B-Chan. Layer3-Parameter          */
#define ISDN_CMD_GETL3  13       /* Get B-Chan. Layer3-Parameter          */
#define ISDN_CMD_LOCK   14       /* Signal usage by upper levels          */
#define ISDN_CMD_UNLOCK 15       /* Release usage-lock                    */

/*
 * Status-Values delivered from lowlevel to linklevel via
 * statcallb().
 *
 */
#define ISDN_STAT_STAVAIL 256    /* Raw status-data available             */
#define ISDN_STAT_ICALL   257    /* Incoming call detected                */
#define ISDN_STAT_RUN     258    /* Signal protocol-code is running       */
#define ISDN_STAT_STOP    259    /* Signal halt of protocol-code          */
#define ISDN_STAT_DCONN   260    /* Signal D-Channel connect              */
#define ISDN_STAT_BCONN   261    /* Signal B-Channel connect              */
#define ISDN_STAT_DHUP    262    /* Signal D-Channel disconnect           */
#define ISDN_STAT_BHUP    263    /* Signal B-Channel disconnect           */
#define ISDN_STAT_CINF    264    /* Charge-Info                           */
#define ISDN_STAT_LOAD    265    /* Signal new lowlevel-driver is loaded  */
#define ISDN_STAT_UNLOAD  266    /* Signal unload of lowlevel-driver      */
#define ISDN_STAT_BSENT   267    /* Signal packet sent                    */
#define ISDN_STAT_NODCH   268    /* Signal no D-Channel                   */
#define ISDN_STAT_ADDCH   269    /* Add more Channels                     */
#define ISDN_STAT_CAUSE   270    /* Cause-Message                         */

/*
 * Values for feature-field of interface-struct.
 */
#define ISDN_FEATURE_L2_X75I    (1 << ISDN_PROTO_L2_X75I)
#define ISDN_FEATURE_L2_X75UI   (1 << ISDN_PROTO_L2_X75UI)
#define ISDN_FEATURE_L2_X75BUI  (1 << ISNN_PROTO_L2_X75BUI)
#define ISDN_FEATURE_L2_HDLC    (1 << ISDN_PROTO_L2_HDLC)

#define ISDN_FEATURE_L3_TRANS   (256 << ISDN_PROTO_L3_TRANS)

/*
 * Structure for exchanging above infos
 *
 */
typedef struct {
  int   driver;                  /* Lowlevel-Driver-ID                    */
  int   command;                 /* Command or Status (see above)         */
  ulong arg;                     /* Additional Data                       */
  char  num[50];                 /* Additional Data                       */
} isdn_ctrl;

/*
 * Function-Table (initialized ad load-time of lowlevel-driver)
 *
 */
typedef struct {
  /* Number of channels supported by this driver
   */
  int channels;
  /* Feature-Flags for this driver.
   * See defines ISDN_FEATURE_... for Values
   */
  
  /* 
   * Maximum Size of transmit/receive-buffer this driver supports.
   */
  int maxbufsize;

  unsigned long features;
  /* Receive-Callback
   * Parameters:
   *             int    Driver-ID
   *             int    local channel-number (0 ...)
   *             u_char pointer to received data (in Kernel-Space, volatile)
   *             int    length of data
   */
  void (*rcvcallb)(int, int, u_char*, int);
  /* Status-Callback
   * Parameters:
   *             isdn_ctrl*
   *                   driver  = Driver ID.
   *                   command = One of above ISDN_STAT_... constants.
   *                   arg     = depending on status-type.
   *                   num     = depending on status-type.
   */
  int (*statcallb)(isdn_ctrl*);
  /* Send command
   * Parameters:
   *             isdn_ctrl*
   *                   driver  = Driver ID.
   *                   command = One of above ISDN_CMD_... constants.
   *                   arg     = depending on command.
   *                   num     = depending on command.
   */
  int (*command)(isdn_ctrl*);
  /* Send Data
   * Parameters:
   *             int    local channel-number (0 ...)
   *             u_char pointer to data (volatile)
   *             int    length of data
   *             int    Flag: 0 = Call form Kernel-Space (use memcpy,
   *                              no schedule allowed) 
   *                          1 = Data is in User-Space (use memcpy_fromfs,
   *                              may schedule)
   */
  int (*writebuf)(int, u_char*, int, int);
  /* Send raw D-Channel-Commands
   * Parameters:
   *             u_char pointer data (volatile)
   *             int    length of data
   *             int    Flag: 0 = Call form Kernel-Space (use memcpy,
   *                              no schedule allowed) 
   *                          1 = Data is in User-Space (use memcpy_fromfs,
   *                              may schedule)
   */
  int (*writecmd)(u_char*, int, int);
  /* Read raw Status replies
   *             u_char pointer data (volatile)
   *             int    length of buffer
   *             int    Flag: 0 = Call form Kernel-Space (use memcpy,
   *                              no schedule allowed) 
   *                          1 = Data is in User-Space (use memcpy_fromfs,
   *                              may schedule)
   */
  int (*readstat)(u_char*, int, int);
} isdn_if;

/*
 * Function which must be called by lowlevel-driver at loadtime with
 * the following fields of above struct set:
 *
 * channels (Number of channels that will be supported)
 * command  Address of Command-Handler
 * features Bitwise coded Features of this driver (use ISDN_FEATURE_...)
 * writebuf Address of Send-Command-Handler
 * writecmd    "    "  D-Channel  " which accepts raw D-Ch-Commands
 * readstat    "    "  D-Channel  " which delivers raw Status-Data
 *
 * The linklevel-driver fills the following fields:
 *
 * channels  Driver-ID assigned to this driver. (Must be used on all
 *           subsequent callbacks.
 * rcvcallb  Address of handler for received data
 * statcallb    "    "     "    for status-changes
 *
 */
int register_isdn(isdn_if*);

#endif /* __KERNEL__ */
#endif /* isdnif_h */















