#include <linux/config.h>
#include <linux/module.h>

#include <linux/mm.h>
#include <linux/malloc.h>
#include <linux/timer.h>

#define byte unsigned char
#include "isdnif.h"
#include "teles.h"

struct IsdnBuffers *tracebuf=NULL;

void BuffersInit(struct IsdnBuffers *ib)
{
  struct IsdnBufHeader *ibh=NULL,*prev;
  int i;
  
  prev=NULL;
  for(i=0;i<ib->bufnr;i++) {
    ibh=(struct IsdnBufHeader *)(ib->bufspace+i*ib->bufsize);
    ibh->next=prev;
    ibh->buf=ib;
    ibh->heldby=NULL;
    ibh->magix=301267;
    prev=ibh;
  }
  ib->freelist=ibh;      
  ib->spacewq=NULL;
  ib->magic=301266;
}

void BuffersNew(struct IsdnBuffers *ib,int size,int nr,
  char *owner)
{
  ib->bufspace=(byte *)Smalloc(size*nr,GFP_KERNEL,owner);
  ib->bufsize=size;
  ib->bufnr=nr;
  ib->spacewq=NULL;
  strcpy(ib->owner,owner);
  
  BuffersInit(ib);
}

void BuffersNewAtom(struct IsdnBuffers *ib,int size,int nr,
  char *owner)
{
  ib->bufspace=(byte *)Smalloc(size*nr,GFP_ATOMIC,owner);
  ib->bufsize=size;
  ib->bufnr=nr;
  ib->spacewq=NULL;
  strcpy(ib->owner,owner);
  
  BuffersInit(ib);
}

#ifdef BUFFERS_DEBUG
static int bufhdrnotok(struct IsdnBufHeader *ibh,char *where)
{
  struct IsdnBuffers *ib;
  int bufnr;
  
  if (ibh->magic!=301268) {
    printk("%s:IsdbBufHeader magic not ok\n",where);
    return(!0);
  }
  
  ib=ibh->buf;

  if (ib->magic!=301266) {
    printk("%s: IsdnBuffers magic not ok\n",where);
    return(!0);
  }
  
  bufnr=(byte *)ibh-ib->bufspace;
  bufnr=bufnr/ib->bufsize;
  if (ib->bufspace+ib->bufsize*bufnr!=(byte *)ibh) {
    printk("%s: not a BufHeader of this buf\n",where);
    return(!0);
  }
  return(0);
}

static int bufnotok(struct IsdnBuffers *ib,char *where)
{
  struct IsdnBufHeader *ibh=ib->freelist;
  
  while (ibh) 
    if (bufhdrnotok(ibh,where)) {
      printk("was in bufnotok\n");
      return(!0);
    }
    else
      ibh=ibh->next;
  return(0);
}
#endif

void BuffersRelease(struct IsdnBuffers *ib)
{
#ifdef BUFFERS_DEBUG
  if (bufnotok(ib,"BuffersRelease")) return;
#endif
  Sfree(ib->bufspace);
}

int ListLength(struct IsdnBufHeader *list)
{
  int i;
  
  i=0;
  while ((list)&&(i<100)) { i++;list=list->next; }
  return(i);
}

int BufGetWait(struct IsdnBufHeader **ibh,
  struct IsdnBuffers *ib,void *heldby,int where)
{

#ifdef BUFFERS_DEBUG
  if (bufnotok(ib,"BufGetWait")) return(-EINTR);
#endif  
  cli();
  while (!0)  
   if (ib->freelist) {
     *ibh=ib->freelist;
     ib->freelist=(*ibh)->next;
     (*ibh)->heldby=heldby;
     (*ibh)->where=where;
     sti();
     return(0);
   }
   else {
     if (current->signal & ~current->blocked) {
       sti();
       return(-EINTR);
     }
     else
       interruptible_sleep_on(&(ib->spacewq));
   }
}

int BufGet(struct IsdnBufHeader **ibh,
  struct IsdnBuffers *ib,void *heldby,int where)
{
  unsigned long flags;
  
#ifdef BUFFERS_DEBUG
  if (bufnotok(ib,"BufGet")) return(-1);
#endif  

  save_flags(flags);cli();
  if (ib->freelist) {
    *ibh=ib->freelist;
    ib->freelist=(*ibh)->next;
    (*ibh)->heldby=heldby;
    (*ibh)->where=where;
    restore_flags(flags);
    return(0);
  }
  restore_flags(flags);
  return(-1);
}

int BufGetInt(struct IsdnBufHeader **ibh,
  struct IsdnBuffers *ib,void *heldby,int where)
{
  return(BufGet(ibh,ib,heldby,where));
}

void BufRelease(struct IsdnBufHeader *ibh)
{
  struct IsdnBuffers *ib=ibh->buf;
  unsigned long flags;

  if (ibh->magix!=301267) {
    printk("BufRelease: illegale buffer\n");
    return;
  }
#ifdef BUFFERS_DEBUG  
  if (bufhdrnotok(ibh,"BufRelease")) return;
#endif
  
  if (ib==tracebuf)
    printk("BufRelease %x\n",(int)ibh);

  save_flags(flags);cli();
  ibh->next=ib->freelist;
  ib->freelist=ibh;
  restore_flags(flags);
  wake_up_interruptible(&(ib->spacewq));
}

void BufReleaseNoInt(struct IsdnBufHeader *ibh)
{
  BufRelease(ibh);
}

#ifdef BUFFERS_DEBUG
static int bqnotok(struct IsdnBufQueue *ibq,char *where)
{
  struct IsdnBufHeader *ibh,*prev;
  
  if (ibq->magic!=301267) {
    printk("%s: IsndBufQueue magic not ok ibq %x\n",where,ibq);
    return(!0);
  }
  
  ibh=ibq->head;
  prev=NULL;
  while (ibh) {

if ( ibh->free!=0 ) printk("Free buffer on this queue\n");
    if (bufhdrnotok(ibh,where)) {
      printk("was in bqnotok\n");
      return(!0);
    }
    else {
      prev=ibh;
      ibh=ibh->next;
    }
}

  if (prev!=ibq->tail) {
    printk("%s: tail %x doesn't match last buf %x\n",where,
	   ibq->tail,prev);
    return(!0);
  }
  
  return(0);
}
#endif


void BufAddQueue(struct IsdnBufQueue *ibq,
  struct IsdnBufHeader *ibh)
{
  unsigned long flags;

#ifdef  BUFFERS_DEBUG
  if (bqnotok(ibq,"BufAddQueue")) return;
  if (bufhdrnotok(ibh,"BufAddQueue")) return;
#endif
  
  save_flags(flags);cli();
  if (!ibq->head) ibq->head=ibh;
  if (ibq->tail) ibq->tail->next=ibh;
  ibq->tail=ibh;
  ibh->next=NULL;
  wake_up_interruptible(&(ibq->wq));
  restore_flags(flags);
}

void BufAddQueueNoInt(struct IsdnBufQueue *ibq,
  struct IsdnBufHeader *ibh)
{
  BufAddQueue(ibq,ibh);
}

int BufRmQueue(struct IsdnBufHeader **ibh,struct IsdnBufQueue *ibq)
{

#ifdef BUFFERS_DEBUG  
  if (bqnotok(ibq,"BufRmQueue")) return(-1);
#endif
  
  cli();

  while (!0)
    if (ibq->head) {
      if (ibq->tail==ibq->head) ibq->tail=NULL;
      *ibh=ibq->head;
      ibq->head=(*ibh)->next;
      sti();
      return(0);
    } 
    else {
      if (current->signal & ~current->blocked) {
        sti();
        return(-EINTR);
      }
      else
        interruptible_sleep_on(&(ibq->wq));
    }               
}

int BufRmQueueInt(struct IsdnBufHeader **ibh,struct IsdnBufQueue *ibq)
{
  long flags;
#ifdef BUFFERS_DEBUG
  if (bqnotok(ibq,"BufRmQueueInt")) return(-1);
#endif
  
  save_flags(flags);cli();
  if (ibq->head) {
    if (ibq->tail==ibq->head) ibq->tail=NULL;
    *ibh=ibq->head;
    ibq->head=(*ibh)->next;
    restore_flags(flags);
    return(0);
  } 
  else {
    restore_flags(flags);
    return(-1);
  }
}

void BufInitQueue(struct IsdnBufQueue *ibq)
{
  ibq->head=NULL;
  ibq->tail=NULL;
  ibq->wq=NULL;
  ibq->magic=301267;
}

void BufReleaseQueue(struct IsdnBufQueue *ibq)
{
  struct IsdnBufHeader *ibh;

  while (ibq->head) {
    BufRmQueue(&ibh,ibq);
    BufRelease(ibh);
  }
}

void Sfree(byte *ptr)
{
#if 0
  printk("Sfree %x\n",ptr);
#endif
  kfree(ptr);
}

byte *Smalloc(int size,int pr,char *why)
{
  byte *p;
  
  p=(byte *)kmalloc(size,pr);
#if 0
  printk("Smalloc %s size %d res %x\n",why,size,p);
#endif
  return(p);
}

void BufRmReleaseNoInt(struct IsdnBufQueue *ibq,int count)
{
  struct IsdnBufHeader *ibh;

  while (count--) {
    BufRmQueueInt(&ibh,ibq);
    BufRelease(ibh);
  }
}

static int bufinlist(struct IsdnBufHeader *list,
  struct IsdnBufHeader *ibh)
{
  while (list)
    if (list==ibh)
      return(!0);
    else
      list=list->next;
  return(0);
}

int wherearemybufs(struct IsdnBuffers *ib)
{
  int i,err;
  struct IsdnBufHeader *ibh;

  err=0;
  for(i=0;i<ib->bufnr;i++) {
    ibh=(struct IsdnBufHeader *)(ib->bufspace+i*ib->bufsize);
    if (!bufinlist(ib->freelist,ibh)) {
      printk("buf %x not in list alloc %d\n",(int)ibh,
        ibh->where);
      err=!0;
    } 
  }
  return(err);
}

void discardq(struct IsdnBufQueue *q,int pr,void *heldby,
  int releasetoo)
{
  long flags;
  struct IsdnBufHeader *sp;

  save_flags(flags);cli();
  		    
  while (!0) {
    sp=q->head;
    if (!sp) break;
    if ((sp->primitive==pr)&&(sp->heldby==heldby)) {
      q->head=sp->next;
      if (q->tail==sp) q->tail=NULL;
      if (releasetoo) BufRelease(sp);
    }
    else
      break;
  }
  
  sp=q->head;
  if (sp)
    while (sp->next) {
      if ((sp->next->primitive==pr)&&(sp->next->heldby==heldby)) {
        if (q->tail==sp->next) q->tail=sp;
        if (releasetoo) BufRelease(sp->next);
        sp->next=sp->next->next;
      }
      else
	sp=sp->next;
    }
  
  restore_flags(flags);
}
